#include "TAnalysisNebula.hh"

#include <iostream>
#include <sstream>
#include <vector>
#include <map>
#include <math.h>

#include <TFile.h>
#include <TH1.h>
#include <TH2.h>
#include <TRandom.h>

#include "TScintillatorLightOutput.hh"
#include "TSimulationDataBeam.hh"
#include "TSimulationDataScintillator.hh"
#include "TSimulationParameterNeutronDetector.hh"

//_____________________________________________________________________________
TAnalysisNebula::TAnalysisNebula()
  : fInputFile(0), fOutputFile(0), fHeader("")
{}
//_____________________________________________________________________________
void TAnalysisNebula::ReadInputFile(TString rootfile){
  fInputFile = new TFile(rootfile.Data(),"readonly");

  std::stringstream ss;

  fInputFile->GetObject("parameterGlobal", fParaGlobal);
  if(!fParaGlobal){
    std::cerr << " fParaGlobal does not exist" << std::endl;
    return;
  }
  ss<<*fParaGlobal<<std::endl;

  fInputFile->GetObject("parameterNEBULA", fParaNebula);
  if(!fParaNebula){
    std::cerr << " fParaNebula does not exist" << std::endl;
    return;
  }
  ss<<*fParaNebula<<std::endl;

  fInputFile->GetObject(fParaGlobal->fTreeName, fTree);
  if(!fTree){
    std::cerr << " cannot find tree" << std::endl;
    return;
  }

  fHeader+=ss.str().c_str();

  fBranchSimulationDataBeamArray = fTree->GetBranch("beam");
  fBranchSimulationDataBeamArray->SetAddress(&gSimulationDataBeamArray);

  fBranchSimulationDataScintillatorArray = fTree->GetBranch("neutronDetector");
  fBranchSimulationDataScintillatorArray->SetAddress(&gSimulationDataScintillatorArray);

}
//_____________________________________________________________________________
void TAnalysisNebula::OpenOutputFile(TString rootfile){

  fOutputFile = new TFile(rootfile.Data(), "recreate");
  if (!fOutputFile){
    std::cerr<< "cannot open "<< rootfile << std::endl;
    return;
  }

}
//_____________________________________________________________________________
void TAnalysisNebula::Analyze(){
  if (! fInputFile){
    std::cerr<< "InputRootFile is not defined, call ReadInputFile"<<std::endl;
    return;
  }

  SetParameters();
  std::cout<<fHeader<<std::endl;

  if (fOutputFile) fOutputFile->cd();

  BookHist();

  // event loop
  Long64_t Nentries = fTree->GetEntries();
  for (int i=0;i<Nentries; i++){
    if (i%1000==0)
      std::cout<< "\r Event: "<<i
	       <<" / "<<Nentries
	       <<" ("<<(double)i/Nentries*100.<<"%)"
	       <<std::flush;

    fBranchSimulationDataBeamArray->GetEntry(i);
    fBranchSimulationDataScintillatorArray->GetEntry(i);
    ProcessOneEvent();
  }

  // store result in root file
  if (fOutputFile) {
    TNamed Header("Header","");
    Header.SetTitle(fHeader.Data());
    Header.Write();

    fParaGlobal->Write();
    fParaNebula->Write();

    fOutputFile->Write();
  }
}
//_____________________________________________________________________________
void TAnalysisNebula::SetParameters(){
  fNeutThre = 6.0; // MeVee
  fVetoThre = 1.0; // MeVee
  fTimeReso = 0.17*sqrt(2.0); // ns
  fV_scinti = 15.8; // cm/ns
  fNeutAttLen = 668;//cm
  fVetoAttLen = 258;//cm

  std::stringstream ss;
  ss <<" fNeutThre = "<< fNeutThre <<" MeVee"<< std::endl
     <<" fVetoThre = "<< fVetoThre <<" MeVee"<< std::endl
     <<" fTimeReso = "<< fTimeReso <<" ns/PMT"<< std::endl
     <<" fV_scinti = "<< fV_scinti <<" cm/ns"<< std::endl
     <<" fNeutAttLen = "<< fNeutAttLen <<" cm"<< std::endl
     <<" fVetoAttLen = "<< fVetoAttLen <<" cm"<< std::endl;

  fHeader += ss.str().c_str();

}
//_____________________________________________________________________________
void TAnalysisNebula::BookHist(){
  fheff = new TH1I("heff","Efficiency",2,-0.5,1.5);
  fhid = new TH1I("hid","ID",144,0.5,144.5);
  fhmulti = new TH1I("hmulti","Multiplicity",10,-0.5,9.5);
  fhedepo = new TH1D("hedepo","Edepo (MeV)",200,0,200);
  fhidlayer = new TH2I("hidlayer","ID Layer",144,0.5,144.5, 2,0.5,2.5);
  fhidsublayer = new TH2I("hidsublayer","ID SubLayer",144,0.5,144.5, 3,-0.5,2.5);
  fhidsublayerseq = new TH2I("hidsublayerseq","ID SubLayerseq",144,0.5,144.5, 5,-0.5,4.5);

  fhidtu = new TH2D("hidtu","ID Tu",144,0.5,144.5, 200,30,130);
  fhidtd = new TH2D("hidtd","ID Td",144,0.5,144.5, 200,30,130);
  fhidqu = new TH2D("hidqu","ID Qu",144,0.5,144.5, 200,0,200);
  fhidqd = new TH2D("hidqd","ID Qd",144,0.5,144.5, 200,0,200);

  fhidt = new TH2D("hidt","ID T",144,0.5,144.5, 400,30,130);
  fhidt_sim = new TH2D("hidt_sim","ID T(sim)",144,0.5,144.5, 400,30,130);
  fhidq = new TH2D("hidq","ID Q",144,0.5,144.5, 200,0,200);

  fhidx = new TH2D("hidx","ID X",144,0.5,144.5, 200,-200,200);
  fhidy = new TH2D("hidy","ID Y",144,0.5,144.5, 200,-200,200);
  fhidz = new TH2D("hidz","ID Z",144,0.5,144.5, 200,1050,1250);

  fhidx_sim = new TH2D("hidx_sim","ID X(sim)",144,0.5,144.5, 200,-200,200);
  fhidy_sim = new TH2D("hidy_sim","ID Y(sim)",144,0.5,144.5, 200,-200,200);
  fhidz_sim = new TH2D("hidz_sim","ID Z(sim)",144,0.5,144.5, 200,1050,1250);

}
//_____________________________________________________________________________
void TAnalysisNebula::ProcessOneEvent(){

  fvdata.clear();
  SimToExp();

  int multi = fvdata.size();
  fhmulti->Fill(multi);
  for (int i=0;i<multi;i++) {

    fhid->Fill(fvdata[i].id);
    fhedepo->Fill(fvdata[i].q);
    fhidlayer->Fill(fvdata[i].id, fvdata[i].layer);
    fhidsublayer->Fill(fvdata[i].id, fvdata[i].sublayer);
    fhidsublayerseq->Fill(fvdata[i].id, fvdata[i].sublayerseq);

    fhidtu->Fill(fvdata[i].id, fvdata[i].tu);
    fhidtd->Fill(fvdata[i].id, fvdata[i].td);
    fhidqu->Fill(fvdata[i].id, fvdata[i].qu);
    fhidqd->Fill(fvdata[i].id, fvdata[i].qd);

    fhidt->Fill(fvdata[i].id, fvdata[i].t);
    fhidt_sim->Fill(fvdata[i].id, fvdata[i].t_sim);
    fhidq->Fill(fvdata[i].id, fvdata[i].q);

    fhidx->Fill(fvdata[i].id, fvdata[i].pos.x());
    fhidy->Fill(fvdata[i].id, fvdata[i].pos.y());
    fhidz->Fill(fvdata[i].id, fvdata[i].pos.z());

    fhidx_sim->Fill(fvdata[i].id, fvdata[i].pos_sim.x());
    fhidy_sim->Fill(fvdata[i].id, fvdata[i].pos_sim.y());
    fhidz_sim->Fill(fvdata[i].id, fvdata[i].pos_sim.z());

  }
  if (multi>0) fheff->Fill(1);
  else         fheff->Fill(0);


  return;
}
//_____________________________________________________________________________
void TAnalysisNebula::SimToExp(){

  int Nhit = gSimulationDataScintillatorArray->size();

  std::map<int,TSimulationParameterNeutronDetector*> *ParaMap = &(fParaNebula->fParaNeutDetecMap);

  for(int ihit=0; ihit<Nhit; ++ihit){
    TSimulationDataScintillator DataScinti = (*gSimulationDataScintillatorArray)[ihit];
    std::vector<Int_t> id = DataScinti.fModuleID;

    TSimulationParameterNeutronDetector *para = (*ParaMap)[id[0]];

    Double_t Xsiz=0;
    Double_t Ysiz=0;
    Double_t Zsiz=0;
    Double_t AttLen=0;
    if       (para->fDetecType == "Neut"){
      Xsiz = fParaNebula->fNeutSize.x();
      Ysiz = fParaNebula->fNeutSize.y();
      Zsiz = fParaNebula->fNeutSize.z();
      AttLen = fNeutAttLen;
    }else if (para->fDetecType == "Veto"){ 
      Xsiz = fParaNebula->fVetoSize.x();
      Ysiz = fParaNebula->fVetoSize.y();
      Zsiz = fParaNebula->fVetoSize.z();
      AttLen = fVetoAttLen;
    }
    Double_t YPosAtDetec = DataScinti.fPrePosition.Y()// detected position in the detecotr frame
      - ( fParaNebula->fPosition.y() + para->fPosition.y() );
    Double_t dy_u = 0.5*Ysiz - YPosAtDetec;// length to PMT
    Double_t dy_d = 0.5*Ysiz + YPosAtDetec;

    Double_t Edepo_MeVee = TScintillatorLightOutput::MeVtoMeVee(DataScinti.fPreKineticEnergy,
								DataScinti.fPreKineticEnergy-
								DataScinti.fEnergyDeposit,
								DataScinti.fParticleName.Data(),
								DataScinti.fPDGCode);

    TAnalysisDataScintillator data;
    data.id    = id[0];
    data.layer = para->fLayer;
    data.sublayer  = para->fSubLayer;
    data.sublayerseq  = para->fSubLayerSeq;

    data.tu    = DataScinti.fPreTime + dy_u/fV_scinti;
    data.td    = DataScinti.fPreTime + dy_d/fV_scinti;
    data.qu    = Edepo_MeVee*exp(-dy_u/AttLen);
    data.qd    = Edepo_MeVee*exp(-dy_d/AttLen);
    data.q_sim = Edepo_MeVee;
    Double_t Xmeas = para->fPosition.X()+fParaNebula->fPosition.X()
      + gRandom->Uniform(-0.5*Xsiz,0.5*Xsiz);
    data.pos.SetX(Xmeas);
    Double_t Zmeas = para->fPosition.Z()+fParaNebula->fPosition.Z()
      + gRandom->Uniform(-0.5*Zsiz,0.5*Zsiz);
    data.pos.SetZ(Zmeas);
    data.pos_sim = DataScinti.fPrePosition;


    // search same ID
    bool IDFound = false;
    for (int j=0;j<(int)fvdata.size(); ++j){
      if (fvdata[j].id == data.id){
	IDFound = true;
	if (fvdata[j].tu > data.tu) fvdata[j].tu = data.tu;
	if (fvdata[j].td > data.td) fvdata[j].td = data.td;
	fvdata[j].qu += data.qu;
	fvdata[j].qd += data.qd;
	if (fvdata[j].tu + fvdata[j].td >
	    data.tu + data.td ) fvdata[j].pos_sim = data.pos_sim;
	break;
      }
    }
    if (! IDFound) fvdata.push_back(data);

  }//for(int ihit=0; ihit<Nhit; ++ihit)



  //-----------------------
  // remove low energy deposit below threshold
  for (int i=0; i<(int)fvdata.size(); ++i){
    Int_t id = fvdata[i].id;
    TSimulationParameterNeutronDetector *para = (*ParaMap)[id];
    Double_t thre = 0;
    if     (para->fDetecType=="Neut")  thre = fNeutThre;
    else if(para->fDetecType=="Veto")  thre = fVetoThre;

    fvdata[i].q = sqrt(fvdata[i].qu * fvdata[i].qd);
    if (fvdata[i].q < thre ) {
      fvdata.erase(fvdata.begin()+i);
      --i;
    }
  }

  //-----------------------
  // calculate averaged observables etc.
  for (int i=0; i<(int)fvdata.size(); ++i){
    Int_t id = fvdata[i].id;
    TSimulationParameterNeutronDetector *para = (*ParaMap)[id];
    Double_t Ysiz=0;
    if      (para->fDetecType == "Neut")  Ysiz = fParaNebula->fNeutSize.y();
    else if (para->fDetecType == "Veto")  Ysiz = fParaNebula->fVetoSize.y();

    fvdata[i].t_sim = 0.5*(fvdata[i].tu + fvdata[i].td - Ysiz/fV_scinti);

    fvdata[i].tu += gRandom->Gaus(0,fTimeReso);
    fvdata[i].td += gRandom->Gaus(0,fTimeReso);
    fvdata[i].t = 0.5*(fvdata[i].tu + fvdata[i].td - Ysiz/fV_scinti);
    fvdata[i].dt = fvdata[i].td - fvdata[i].tu;
    fvdata[i].pos.SetY(0.5*fvdata[i].dt*fV_scinti + para->fPosition.y() + fParaNebula->fPosition.y());
  }


}
//_____________________________________________________________________________
