
#include "PrimaryGeneratorActionSequentialDecay.hh"
#include "ActionBasicMessenger.hh"
#include "ActionSequentialDecayMessenger.hh"

#include "G4Event.hh"
#include "G4ParticleTable.hh"
#include "G4ParticleDefinition.hh"
#include "G4ParticleGun.hh"

#include "G4PrimaryVertex.hh"
#include "G4PrimaryParticle.hh"

#include "G4RandomDirection.hh"
#include "Randomize.hh"
#include "G4LorentzVector.hh"
#include "G4Tubs.hh"
#include "G4LogicalVolume.hh"
#include "G4LogicalVolumeStore.hh"

#include "TSimulationDataBeam.hh"
#include "TSimulationParameterGlobal.hh"

#include "TRandom3.h"
#include "TVector3.h"
#include "TGenPhaseSpace.h"

PrimaryGeneratorActionSequentialDecay::PrimaryGeneratorActionSequentialDecay(G4int seed)
  : PrimaryGeneratorActionBasic(seed)
{
  fRandom0 = new TRandom3(seed);
  fRandom1 = new TRandom3(seed);
  fOKInitSequentialDecay = false;

  if(!SetBeamType("MonoEnergy")) G4Exception(0, 0, FatalException, 0);

  fActionSequentialDecayMessenger = ActionSequentialDecayMessenger::GetInstance();
  fActionSequentialDecayMessenger->Register(this);
}

PrimaryGeneratorActionSequentialDecay::~PrimaryGeneratorActionSequentialDecay()
{
  // it is done in PrimaryGeneratorActionBasic
//  delete fParticleGun;
}

bool PrimaryGeneratorActionSequentialDecay::SetBeamType(G4String beamType)
{
  if(beamType == "Geantino"){
    beamFunctionPtr = &PrimaryGeneratorActionSequentialDecay::BeamTypeGeantino;
  }else if(beamType == "MonoEnergy"){
    beamFunctionPtr = &PrimaryGeneratorActionSequentialDecay::BeamTypeMonoEnergy;
  }else if(beamType == "MonoEnergyRandTheta"){
    beamFunctionPtr = &PrimaryGeneratorActionSequentialDecay::BeamTypeMonoEnergyRandTheta;
  }else if(beamType == "MonoEnergyRandMomentum"){
    beamFunctionPtr = &PrimaryGeneratorActionSequentialDecay::BeamTypeMonoEnergyRandMomentum;
  }else if(beamType == "MonoEnergyRandMomentumRandPosition"){
    beamFunctionPtr = &PrimaryGeneratorActionSequentialDecay::BeamTypeMonoEnergyRandMomentumRandPosition;
  }else if(beamType == "GausEnergy"){
    beamFunctionPtr = &PrimaryGeneratorActionSequentialDecay::BeamTypeGausEnergy;
  }else if(beamType == "GausEnergyRandTheta"){
    beamFunctionPtr = &PrimaryGeneratorActionSequentialDecay::BeamTypeGausEnergyRandTheta;
  }else if(beamType == "ExpEnergyRandMomentum"){
    beamFunctionPtr = &PrimaryGeneratorActionSequentialDecay::BeamTypeExpEnergyRandMomentum;
  }else if(beamType == "Cs137"){
    beamFunctionPtr = &PrimaryGeneratorActionSequentialDecay::BeamTypeCs137;
  }else if(beamType == "Co60"){
    beamFunctionPtr = &PrimaryGeneratorActionSequentialDecay::BeamTypeCo60;
  }else if(beamType == "Y88"){
    beamFunctionPtr = &PrimaryGeneratorActionSequentialDecay::BeamTypeY88;
  }else if(beamType == "BeamInducedGamma"){
    beamFunctionPtr = &PrimaryGeneratorActionSequentialDecay::BeamTypeBeamInducedGamma;
  }else if(beamType == "DemocraticDecay"){
    beamFunctionPtr = &PrimaryGeneratorActionSequentialDecay::BeamTypeDemocraticDecay;
  }else if(beamType == "SequentialDecay"){
    beamFunctionPtr = &PrimaryGeneratorActionSequentialDecay::BeamTypeSequentialDecay;
  }else{
    G4cerr << "\n !!! beamType \"" << beamType << "\" does not exist !!!" << G4endl;
    return false;
  }
  G4cout << " set beam type to \"" << beamType << "\"" << G4endl;

  // save beam type
  if(!gSimulationParameterGlobal){
    gSimulationParameterGlobal = new TSimulationParameterGlobal;
  }
  gSimulationParameterGlobal->fBeamType = beamType;

  return true;
}

void PrimaryGeneratorActionSequentialDecay::GeneratePrimaries(G4Event* anEvent)
{
  (this->*beamFunctionPtr)(anEvent);
}

void PrimaryGeneratorActionSequentialDecay::AccumulateBeamData(G4Event* anEvent)
{
  if(!gSimulationDataBeamArray) return;
  gSimulationDataBeamArray->clear();
  for(int i=0; i<anEvent->GetNumberOfPrimaryVertex(); ++i){
    G4PrimaryVertex* primaryVertex = anEvent->GetPrimaryVertex(i);
    for(int j=0; j<primaryVertex->GetNumberOfParticle(); ++j){
      G4int primaryParticleID = j+i*j;

      G4PrimaryParticle* primaryParticle = primaryVertex->GetPrimary(j);
      Double_t mass = primaryParticle->GetMass();
      Double_t charge = primaryParticle->GetCharge();
      Int_t pdgCode= primaryParticle->GetPDGcode();
      TVector3 momentum(primaryParticle->GetMomentum().x(),
			primaryParticle->GetMomentum().y(),
			primaryParticle->GetMomentum().z());
      TVector3 position(primaryVertex->GetX0(),
			primaryVertex->GetY0(),
			primaryVertex->GetZ0());
      Double_t time = primaryVertex->GetT0();

      TSimulationDataBeam simulationDataBeam;
      simulationDataBeam.fPrimaryParticleID = primaryParticleID;
      simulationDataBeam.fPDGCode = pdgCode;
      simulationDataBeam.fMass = mass/MeV;
      simulationDataBeam.fCharge = charge/eplus;
      simulationDataBeam.fMomentum = 1/MeV*momentum;
      simulationDataBeam.fPosition = 1/cm*position;
      simulationDataBeam.fTime = time/ns;
      gSimulationDataBeamArray->push_back(simulationDataBeam);

#ifdef DEBUG
      std::cout << simulationDataBeam << std::endl;
#endif
    }
  }
}

//_____________________________________________________________
// Sequential Decay
void PrimaryGeneratorActionSequentialDecay::BeamTypeSequentialDecay(G4Event* anEvent)
{
  if(!fOKInitSequentialDecay) InitSequentialDecay();

  // 1st decay (like 28O,27O -> 26O+2n,1n)
  double weight0;
  weight0 = fPhaseSpace0->Generate();
  while(fRandom0->Uniform()>weight0) weight0 = fPhaseSpace0->Generate();
  TLorentzVector P_int[3]; // momenta of intermediate state

  for(int i=0;i<fNdecay-2;++i)  P_int[i] = *(fPhaseSpace0->GetDecay(i));
  for (int i=3;i<fNdecay;++i) fP[i] = &(P_int[i-2]);
//  fP[3] = &(P_int[1]);
//  fP[4] = &(P_int[2]);

  // 2nd decay (like 26O->24O+2n)
  TGenPhaseSpace PhaseSpace1;
  PhaseSpace1.SetDecay(P_int[0],3,fMasses);
  double weight1;
  weight1 = PhaseSpace1.Generate();
  while(fRandom1->Uniform()>weight1) weight1 = PhaseSpace1.Generate();
  for(int i=0;i<3;++i) fP[i] = PhaseSpace1.GetDecay(i);

  // neutron
  for(int i=1;i<fNdecay;++i){
    fParticleGun->SetParticleDefinition(fParticles[i]);
    fParticleGun->SetParticlePosition(G4ThreeVector(0.*cm,0.*cm,0.*cm));// should be modified
    G4ThreeVector momDir(fP[i]->Px(), fP[i]->Py(), fP[i]->Pz());

    fParticleGun->SetParticleMomentumDirection(momDir);
    G4double KineticEnergy = (fP[i]->E()-fP[i]->M())*MeV;
    fParticleGun->SetParticleEnergy(KineticEnergy);
    fParticleGun->GeneratePrimaryVertex(anEvent);
  }
  AccumulateBeamData(anEvent);

  // core
  {
    int i=0;
    G4ParticleDefinition* particle = fParticles[i];

    Double_t mass = particle->GetPDGMass();
    Double_t charge = particle->GetPDGCharge();
    Int_t pdgCode= particle->GetPDGEncoding();
    TVector3 momentum(fP[i]->Px(), fP[i]->Py(), fP[i]->Pz());
    TVector3 position(0, 0, 0);
    Double_t time = 0;

    TSimulationDataBeam simulationDataBeam;
    simulationDataBeam.fPrimaryParticleID = -1;
    simulationDataBeam.fPDGCode = pdgCode;
    simulationDataBeam.fMass = mass/MeV;
    simulationDataBeam.fCharge = charge/eplus;
    simulationDataBeam.fMomentum = 1/MeV*momentum;
    simulationDataBeam.fPosition = 1/cm*position;
    simulationDataBeam.fTime = time/ns;
    gSimulationDataBeamArray->push_back(simulationDataBeam);
  }
}
//________________________________________________________________
bool PrimaryGeneratorActionSequentialDecay::SetErel012(G4double erel)
{
  fErel012 = erel;
  fOKInitSequentialDecay=false;
  return true;
}
//________________________________________________________________
bool PrimaryGeneratorActionSequentialDecay::InitSequentialDecay()
{
  G4cout<<"InitSequentialDecay :"
	<<" Erel="<<fErel/MeV<<"MeV"
	<<" Erel0123="<<fErel012/MeV<<"MeV"
	<<G4endl;

  fNdecay = fAbeam - fAcore + 1;
  if(fNdecay>NDECAY_MAX) G4cout<< "Ndecay exceed max="<<NDECAY_MAX <<G4endl;
  if(fNdecay<3) G4cout<< "Ndecay="<<fNdecay<<" must be larger than 2"<<G4endl;
  if(fZbeam != fZcore)   G4cout<< "Zbeam - Zcore is not zero!" <<G4endl;

  G4ParticleTable* particleTable = G4ParticleTable::GetParticleTable();
//  G4ParticleDefinition *beam = particleTable->GetIon(fZbeam, fAbeam, 0.0);
  fParticles[0] = particleTable->GetIon(fZcore, fAcore, 0.0);
  for(int i=1;i<fNdecay;++i)
    fParticles[i] = particleTable->FindParticle("neutron");

  double Masses_int[NDECAY_MAX];// masses for intermediate states
  G4double Mbeam = 0;
  for(int i=0;i<fNdecay;++i){
    fMasses[i] = fParticles[i]->GetPDGMass()/MeV;
    Mbeam += fMasses[i]*MeV;
  }
  Masses_int[0] = fMasses[0] + fMasses[1] + fMasses[2] + fErel012;
  for (int i=3;i<fNdecay;++i)  Masses_int[i-2] = fMasses[i];
//    Masses_int[1] = fMasses[3];
//    Masses_int[2] = fMasses[4];

  Mbeam += fErel;
  G4double Tbeam = fBeamEnergy*fAbeam;// Total Kinetic energy
  G4double Ebeam = Tbeam + Mbeam;
  G4double Pbeam_abs = sqrt(Tbeam*(Tbeam+2.0*Mbeam));
  TLorentzVector Pbeam(0.0, 0.0, Pbeam_abs/MeV, Ebeam/MeV);// no angular spread, should be mod.

//  G4cout<<"Mass of Beam = "<<Mbeam/MeV<<G4endl;
//  G4cout<<"core mass = "<<fParticles[0]->GetPDGMass()/MeV<<G4endl;
//  G4cout<<"neutron mass ="<<fParticles[1]->GetPDGMass()/MeV<<G4endl;

  //delete fPhaseSpace;
  fPhaseSpace0 = new TGenPhaseSpace;
  //fPhaseSpace0->SetDecay(Pbeam,3,Masses_int);
  fPhaseSpace0->SetDecay(Pbeam,fNdecay-2,Masses_int);

  fOKInitSequentialDecay=true;
  return true;
}
